// code to compute band structure of 2D periodic potential using plane waves

#include <iostream>
#include <iomanip>
#include <fstream>
#include <complex>
#include <vector>
#include <assert.h>
#include <math.h>
#include "matrix.h"
#ifndef NO_MPI
    #include <mpi.h>
#endif

//physics parameters
#define NUMK    200
#define V0      0.
#define V1      0.
#define NWAVE   3
#define NORB    (2*NWAVE+1)*(2*NWAVE+1) // standard matrix size

using namespace std;

typedef complex<double> cdouble;
typedef matrix<cdouble> cmat;
cdouble II(0,1);



// LAPACK stuff
extern "C" void zheev_(char* jobz, char* uplo, int* N, cdouble* H, int* LDA, double* W, cdouble* work, int* lwork, double* rwork, int *info);

int     matsize = NORB;
int     lwork = 2*(2*NORB-1);
double  rwork[3*NORB-2];
cdouble work[2*(2*NORB-1)];
char    jobz = 'V';
char    uplo = 'U';
int     info;


void diagonalize(cmat &H, vector<double> &evals)
{
	zheev_(&jobz, &uplo, &matsize, H.ptr(), &matsize, &evals[0], &work[0], &lwork, &rwork[0], &info);
	assert(!info);
}

void times(cmat &A, cmat &B, cmat &C)
{
    for(int i=0; i<NORB; ++i)
        for(int j=0; j<NORB; ++j)
        {
            C(i,j) = 0.;
            for(int k=0; k<NORB; ++k)
            {
                C(i,j) += A(i,k)*B(k,j);
            }
        }
}

void times_dn(cmat &A, cmat &B, cmat &C)
{
	for(int i=0; i<NORB; ++i)
        for(int j=0; j<NORB; ++j)
        {
            C(i,j) = 0.;
            for(int k=0; k<NORB; ++k)
            {
                C(i,j) += conj(A(k,i))*B(k,j);
            }
        }
}

void times_nd(cmat &A, cmat &B, cmat &C)
{
	for(int i=0; i<NORB; ++i)
        for(int j=0; j<NORB; ++j)
        {
            C(i,j) = 0.;
            for(int k=0; k<NORB; ++k)
            {
                C(i,j) += A(i,k)*conj(B(j,k));
            }
        }
}

inline double get_ek(double kx, double ky)
{
	return (kx*kx + ky*ky);
}


int main(int argc, char * argv[])
{
    //************** MPI INIT ***************************
  	int numprocs=1, myrank=0, namelen;
    
#ifndef NO_MPI
  	char processor_name[MPI_MAX_PROCESSOR_NAME];
  	MPI_Init(&argc, &argv);
  	MPI_Comm_size(MPI_COMM_WORLD, &numprocs);
  	MPI_Comm_rank(MPI_COMM_WORLD, &myrank);
  	MPI_Get_processor_name(processor_name, &namelen);
    
	//cout << "Process " << myrank << " on " << processor_name << " out of " << numprocs << " says hello." << endl;
	MPI_Barrier(MPI_COMM_WORLD);
    
	MPI_Datatype ttype  = MPI_DOUBLE;
    
#endif
	if(myrank==0) cout << "\n\tProgram running on " << numprocs << " processors." << endl;
    
	//if(myrank==0)
	//	printf("\n\n");
    
#ifndef NO_MPI
	MPI_Barrier(MPI_COMM_WORLD);
#endif
    
    /*
	if(myrank==0)
    {
		cout << endl;
        cout << setprecision(50);
        cout << M_PI << endl;
        cout << setprecision(6);
    }
     */
    
    ofstream outfile;
    
    if(myrank==0)
    {
        outfile.open("bands.dat");
        outfile << setprecision(8);
    }
 
    
    cmat Hmat(NORB, NORB);
    vector<double> evals(NORB);
 
    int kcounter = 0;
    
    for(double ky=0; ky<=M_PI; ky+=M_PI/NUMK)
    {
        double kx = 0.;
        
        // walk through reciprocal lattice
        // diagonal elements
        for(int ix=-NWAVE; ix<=NWAVE; ++ix)
            for(int iy=-NWAVE; iy<=NWAVE; ++iy)
            {
                Hmat((ix+NWAVE) + (2*NWAVE+1)*(iy+NWAVE), (ix+NWAVE) + (2*NWAVE+1)*(iy+NWAVE)) = get_ek(kx+ix*2.*M_PI, ky+iy*2.*M_PI);
            }
        
        // off-diagonal elements for V potential
        for(int ix=-NWAVE; ix<=NWAVE; ++ix)
            for(int iy=-NWAVE; iy<=NWAVE; ++iy)
                for(int ix2=-NWAVE; ix2<=NWAVE; ++ix2)
                    for(int iy2=-NWAVE; iy2<=NWAVE; ++iy2)
                    {
                        if((ix==ix2 && abs(iy-iy2)==1) || (iy==iy2 && abs(ix-ix2)==1))
                            Hmat((ix+NWAVE) + (2*NWAVE+1)*(iy+NWAVE), (ix2+NWAVE) + (2*NWAVE+1)*(iy2+NWAVE)) = V0;
                        
                        if(abs(ix-ix2)==1 && abs(iy-iy2)==1)
                            Hmat((ix+NWAVE) + (2*NWAVE+1)*(iy+NWAVE), (ix2+NWAVE) + (2*NWAVE+1)*(iy2+NWAVE)) = V1;
                        
                    }
        
        
        diagonalize(Hmat, evals);
        
        if(myrank==0)
        {
            outfile << kcounter << '\t' << kx << '\t' << ky;
            for(int i=0; i<NORB; ++i)
            {
                outfile << '\t' << evals[i];
            }
            outfile << endl;
        }
        kcounter ++;
        
        
    }
    
    for(double kx=0; kx<=M_PI; kx+=M_PI/NUMK)
    {
        double ky = M_PI;
        
        // walk through reciprocal lattice
        // diagonal elements
        for(int ix=-NWAVE; ix<=NWAVE; ++ix)
            for(int iy=-NWAVE; iy<=NWAVE; ++iy)
            {
                Hmat((ix+NWAVE) + (2*NWAVE+1)*(iy+NWAVE), (ix+NWAVE) + (2*NWAVE+1)*(iy+NWAVE)) = get_ek(kx+ix*2.*M_PI, ky+iy*2.*M_PI);
            }
        
        // off-diagonal elements for V potential
        for(int ix=-NWAVE; ix<=NWAVE; ++ix)
            for(int iy=-NWAVE; iy<=NWAVE; ++iy)
                for(int ix2=-NWAVE; ix2<=NWAVE; ++ix2)
                    for(int iy2=-NWAVE; iy2<=NWAVE; ++iy2)
                    {
                        if((ix==ix2 && abs(iy-iy2)==1) || (iy==iy2 && abs(ix-ix2)==1))
                            Hmat((ix+NWAVE) + (2*NWAVE+1)*(iy+NWAVE), (ix2+NWAVE) + (2*NWAVE+1)*(iy2+NWAVE)) = V0;
                        
                        if(abs(ix-ix2)==1 && abs(iy-iy2)==1)
                            Hmat((ix+NWAVE) + (2*NWAVE+1)*(iy+NWAVE), (ix2+NWAVE) + (2*NWAVE+1)*(iy2+NWAVE)) = V1;
                        
                    }

        
        diagonalize(Hmat, evals);
        
        if(myrank==0)
        {
            outfile << kcounter << '\t' << kx << '\t' << ky;
            for(int i=0; i<NORB; ++i)
            {
                outfile << '\t' << evals[i];
            }
            outfile << endl;
        }
        kcounter ++;
        
        
    }
    
    for(double kx=M_PI; kx>=0; kx-=M_PI/NUMK)
    {
        
        double ky = kx;
        
        // walk through reciprocal lattice
        // diagonal elements
        for(int ix=-NWAVE; ix<=NWAVE; ++ix)
            for(int iy=-NWAVE; iy<=NWAVE; ++iy)
            {
                Hmat((ix+NWAVE) + (2*NWAVE+1)*(iy+NWAVE), (ix+NWAVE) + (2*NWAVE+1)*(iy+NWAVE)) = get_ek(kx+ix*2.*M_PI, ky+iy*2.*M_PI);
            }
        
        // off-diagonal elements for V potential
        for(int ix=-NWAVE; ix<=NWAVE; ++ix)
            for(int iy=-NWAVE; iy<=NWAVE; ++iy)
                for(int ix2=-NWAVE; ix2<=NWAVE; ++ix2)
                    for(int iy2=-NWAVE; iy2<=NWAVE; ++iy2)
                    {
                        if((ix==ix2 && abs(iy-iy2)==1) || (iy==iy2 && abs(ix-ix2)==1))
                            Hmat((ix+NWAVE) + (2*NWAVE+1)*(iy+NWAVE), (ix2+NWAVE) + (2*NWAVE+1)*(iy2+NWAVE)) = V0;
                        
                        if(abs(ix-ix2)==1 && abs(iy-iy2)==1)
                            Hmat((ix+NWAVE) + (2*NWAVE+1)*(iy+NWAVE), (ix2+NWAVE) + (2*NWAVE+1)*(iy2+NWAVE)) = V1;
                        
                    }
        
        
        diagonalize(Hmat, evals);
        
        if(myrank==0)
        {
            outfile << kcounter << '\t' << kx << '\t' << ky;
            for(int i=0; i<NORB; ++i)
            {
                outfile << '\t' << evals[i];
            }
            outfile << endl;
        }
        kcounter ++;
        
        
    }
    


    
    if(myrank==0)
        cout << "\n ... done! \n"<<flush;
    
    
#ifndef NO_MPI
	MPI_Finalize();
#endif
    
    return 0;

}


